# Implement C-style function prototypes
# 2000-01-07 unslider

package Proto;
use Exporter;
@ISA=qw(Exporter);
@EXPORT=qw(proto);

# Check if arguments match a prototype. If so, drop the args into the
# appropriate sub-local variables and call the sub; if not, whine.
# Arguments: info on the caller (as an array ref), the name of the sub, the
# sub itself (as a code ref), the prototype (as a string) and the sub's
# argument list (as an array ref)
sub tryproto {
	my($caller,$subname,$sub,$proto,$args)=@_;
	my($package,$filename,$line)=@$caller;
	my $template;
	my %oldsyms;
	my $varargs=0;	# set to 1 if the prototype ends with "..."
	my($foo, $bar);
# we want to make the called sub act like it has a 'my' at the top with all the
# arguments. to do this, we shit refs to the arguments into the symbol table.
# every time we munge a symbol, we store the old symbol name and a ref to its
# values in %oldsyms. then, at completion, we restore the old symbols.

	foreach $template (split(/,/,$proto)) {
		for($template) {
		/^...$/ && do {
			$varargs++;
			last;
		};
		/^\$(.+)/ && do {
			die "not enough arguments to '$subname' in '$filename':$line\n" if(!defined($$args[0]));
			die "invalid argument to '$subname' in '$filename':$line\n" if(ref($$args[0]));
			$foo=$package."::$1";
			$oldsyms{$foo}=\${$foo};
			*{$foo}=\$$args[0];
			shift @$args;
			last;
		};
		# we could put code to grok passing of refs here, or
		# typechecking that a scalar contains only digits, or...
		default:
			die "invalid template '$template' for '$subname'\n";
		}
	}

	if(scalar(@$args)!=0 && !$varargs) {
		die "Too many arguments to '$subname' in '$filename':$line\n";
	}

	&$sub(@$args);

	foreach $foo (keys %oldsyms) {
		*{$foo}=$oldsyms{$foo};
	}
}

# Given a sub and a prototype, generate a new sub that checks the arguments
# against the prototype and then calls the original sub. Replace the original
# sub with the new sub.
sub proto {
	my($subname,$proto)=@_;
	my @caller=caller();
	my $package=$caller[0];

	$subname=$package."::$subname";
	$oldsub=\&{$subname};

	$newsub=sub {
		tryproto(\@caller,$subname,$oldsub,$proto,\@_);
	};

	*{$subname}=$newsub;
}

1;
